# Copyright © 2019-2025 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# ALXmail extends ALXhtml and provides a convenient interface for sending
# emails using parameters configured in alx.ini or passed at runtime.
#
# Attachments and html/plain formats are supported as well as multiple
# send attempts

import os.path
import socket
from time import sleep
import re
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.image import MIMEImage
from smtplib import SMTPException, SMTPAuthenticationError, SMTPConnectError, SMTPDataError, SMTPHeloError, \
    SMTPNotSupportedError, SMTPRecipientsRefused, SMTPResponseException, SMTPSenderRefused, SMTPServerDisconnected
import mimetypes
from alx.html import ALXhtml
import smtplib
from email.message import EmailMessage, Message
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Union


class ALXmail(ALXhtml):
    def __init__(self, mail_type: str = "html") -> None:
        """
        Class to send itrs_email - both text and HTML (default). It is a subclass
        of `alx.html.ALXhtml` to allow simple HTML mail composition. Configuration
        is stored in `alx.ini`

        :param type: Either 'text' or 'html' (default)
        :return: None
        """
        super().__init__()

        self.mail_type = mail_type.lower()
        """The type of the email - 'text' or 'html'. Default is html"""

        if mail_type == "plain":
            self.body = ""
            self.message = EmailMessage()
            """The message as a plain or MIME multipart instance"""
        elif mail_type == "html":
            self.message = MIMEMultipart()
        else:
            raise TypeError("Only 'plain' and 'html' are supported")

        self.sender = self.config.get("mail", "from")
        """The email sender"""
        self.subject = "No subject"
        """The email subject"""
        self.server = None
        """The smtp server object"""
        self.images = 0
        """Used for naming multiple image attachments"""
        self.mailhost = self.config.get("mail", "server")
        """The hostname of the smtp server. Configured in alx.ini"""
        self.recipients = []
        """A list of recipients"""
        self.cc = []
        """A list of those on the cc list"""
        self.bcc = []
        """A list of those on the bcc list"""
        self.attachments = []
        """A list of attachments"""

    def set_from(self, sender: str) -> None:
        """
        Set the address and name of sender
        :param sender: The name and email address of the sender: 'First Last <sender@example.com>'
        """
        self.sender = sender

    def set_subject(self, subject: str) -> None:
        """
        Sets the subject of the message
        :param subject: The subject
        """
        self.subject = subject

    def add_recipient(self, to: str, recipient_type: str = "to") -> None:
        """
        Adds a recipient of the message to the list of recipients
        :param to: a single email address
        :param recipient_type: The recipent type: 'to', 'cc', or 'bcc'
        """
        if recipient_type == "to":
            self.recipients.append(to)
        elif recipient_type == "cc":
            self.cc.append(to)
        elif recipient_type == "bcc":
            self.bcc.append(to)

    def set_recipients(self, to: Union[str, list], recipient_type="to") -> None:
        """
        Sets the list of recipients for the email
        :param to: A list or string of recipients, separated by ' ', ',' or ';'
        :param recipient_type: The recipent type: 'to', 'cc', or 'bcc'. This
         parameter is generally not needed as the helper functions should be used
        """

        if type(to) is list:
            recipients = to
        else:
            recipients = re.split(r"[,\s;]+", to)

        for to in recipients:
            self.add_recipient(to, recipient_type)

    def add_cc(self, cc: str) -> None:
        """
        Adds a cc recipient or list of recipients

        :param cc: A list or string of recipients, separated by ' ', ',' or ';'
        """
        self.set_recipients(cc, "cc")

    def add_bcc(self, bcc: str) -> None:
        """
        Adds a bcc recipient or list of recipients

        :param bcc: A list or string of recipients, separated by ' ', ',' or ';'
        """
        self.set_recipients(bcc, "bcc")

    def set_body(self, body: str) -> None:
        """
        Sets the body of the message to the parameter

        :param body: The body of the message
        """
        self.body = body

    def add_paragraph(self, p: str) -> None:
        """
        Adds a paragraph to the email in the appropriate format.  Either 'plain'
        or 'html'
        :param p: The text for the paragraph
        """
        if self.mail_type == "html":
            super().add_paragraph(p)
        else:
            self.body += "\n" + p + "\n"

    def add_text(self, t: str) -> None:
        """
        Wrapper for `add_paragraph` for backwards compatability

        :param t: The text to add
        """
        self.add_paragraph(t)

    def add_attachment(self, filename: str) -> str:
        """
        Adds a file as an attachment. If an image, it will be displayed
        in the location it is attached. MIME types of application and
        audio are also attached with the file type in mind. It could be
        called as follows to put an image in a table cell created with
        `add_table` and `add_cell`:
        ```
            cell = ("<a href=%s>%s</a><a href='%s'>%s</a>" %
                 (url, k, url, mail.add_attachment(filename)))
            mail.add_cell(cell)
        ```

        :param filename: the path to the file to attach.

        :return: The Content-ID of the attachment
        """

        # If you are composing a plain text email, do not try to
        # add an attachment. Tell your manager to fire you and
        # find someone who knows what they are doing
        assert (self.mail_type == 'html')

        if not os.path.isfile(filename):
            raise FileNotFoundError(filename)

        cd = 'attachment'   # Content-Disposition
        content_id = ""
        fn = os.path.basename(filename)
        with open(filename, "rb") as fp:
            data = fp.read()

        maintype, subtype = mimetypes.guess_type(filename)
        if maintype and maintype.startswith('image'):
            attachment = MIMEImage(data)
            if self.mail_type == 'html':
                cd = 'inline'
                image = 'image%02d' % self.images
                # The content_id is returned so the application
                # can put it where it wants
                content_id = '<img src="cid:%s" width="100%%">' % image
                self.images += 1
                attachment.add_header('Content-ID', '<%s>' % image)
        elif maintype and maintype.startswith('application'):
            attachment = MIMEApplication(data)
        elif maintype and maintype.startswith('audio'):
            attachment = MIMEAudio(data)
        else:
            data = data.decode()
            attachment = MIMEText(data)

        attachment.add_header('Content-Disposition', cd,
                              filename=fn)

        self.attachments.append(attachment)

        return content_id

    def send(self) -> None:
        """
        Sends the message. It firsts constructs the email message
        from the all the parts provided and adds the attachments.
        It also tries a number of times (20) with a 5-second
        delay between each attempt - in case there is a network
        or mailhost glitch. If the number of attempts is exceeded,
        the corresponding exception is thrown. The SMTP host to use is
        configured in `alx.ini`. As are retries and delay.
        """
        self.message["From"] = self.sender
        self.message["Subject"] = self.subject
        # Remove duplicate names
        self.recipients = list(set(self.recipients))
        self.message["To"] = ", ".join(self.recipients)
        if len(self.cc) > 0:
            self.cc = list(set(self.cc))
            self.message["Cc"] = ", ".join(self.cc)
        if len(self.bcc) > 0:
            self.bcc = list(set(self.bcc))
            self.message["Bcc"] = ", ".join(self.bcc)

        if self.mail_type == "html":
            for a in self.attachments:
                self.message.attach(a)

        if self.mail_type == "html":
            body = self.get_html()
            part = MIMEText(body, "html")
            self.message.attach(part)
        else:
            body = self.body + "\n"
            self.message.set_content(body)

        retries = self.config.getint("mail", "retries")
        delay = self.config.getint("mail", "delay")
        count = 0

        while count < retries:
            try:
                self.server = smtplib.SMTP(self.mailhost)
                self.server.send_message(self.message)
                self.server.quit()
                return
            except (socket.error, SMTPException, SMTPAuthenticationError,
                    SMTPConnectError, SMTPDataError, SMTPHeloError,
                    SMTPNotSupportedError, SMTPRecipientsRefused,
                    SMTPResponseException, SMTPSenderRefused,
                    SMTPServerDisconnected) as ex:
                count += 1
                if count == loop:
                    raise type(ex).__name__(format(ex))
                else:
                    sleep(delay)

    def _get_mime_message(self) -> Message:
        """
        Builds and returns the MIME message without sending it.
        Used for testing and inspection.

        :return: The email message
        """
        if self.mail_type == "plain":
            from email.message import EmailMessage
            msg = EmailMessage()
            msg.set_content(self.body)
        else:
            msg = MIMEMultipart()
            for a in self.attachments:
                msg.attach(a)
            msg.attach(MIMEText(self.get_html(), "html"))

        msg["From"] = self.sender
        msg["To"] = ", ".join(self.recipients)
        msg["Subject"] = self.subject
        if self.cc:
            msg["Cc"] = ", ".join(self.cc)
        if self.bcc:
            msg["Bcc"] = ", ".join(self.bcc)

        return msg



# Copyright © 2019 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# Provides a consistent framework for creating itrs toolkit samplers
# without the need to know internal details

from collections import OrderedDict
import sys


class Toolkit:
    def __init__(self, filename: str = None, display_on_exit: bool = True):
        """
        Populate and output an [ITRS geneos toolkit sampler](https://docs.itrsgroup.com/docs/geneos/7.2.0/collection/toolkit-plugin).

        :param filename: If a filename is set then the output goes to
        this file rather than the default of `sys.stdout`
        :param display_on_exit: When exiting the sampler with a call to
        `ok`, `warning` or `error` should the contents
        of the sampler be displayed?  Default is True
        """
        self.headlines = OrderedDict()
        """The headlines of the toolkit sampler, added using
        `add_headline`"""
        self.rows = []
        """A list of each row"""
        self.num_rows = 0
        """The number of rows - incremented as each row is added"""
        self.headings = []
        """A list of headings for the sampler"""
        self._num_columns = 0
        self._filename = filename
        self._display_on_exit = display_on_exit

    def add_headline(self, label: str, value):
        """
        Add a headline to the toolkit
        :param label: The label for the headline
        :param value: the value for the headline
        """
        self.headlines[label] = str(value).replace(",", "\\,")

    def add_headings(self, headings: str, delimiter: str = ','):
        """
        Add a list of headings to the sampler.  Each heading in the list
        is added with a call to `self.add_heading`

        :param headings: delimited string of headings
        :param delimiter: the delimiter (default ',')
        """
        for h in headings.split(delimiter):
            self.add_heading(h)

    def add_heading(self, heading: str):
        """
        Add a single column heading to the sampler
        :param heading: The heading string
        """
        heading = str(heading).strip()
        heading.replace(",", "\\,")
        self.headings.append(heading)
        self._num_columns += 1

    def add_row(self, values, delimiter=','):
        """
        Add a row to the sampler.  A list of values or a delimited string
        can be passed.  Handles quoting of values to ensure correct display

        :param values: A list of values or a delimited string
        :param delimiter: If a string is passed then each column is delimited
        by this character.  Default is ','
        """
        if isinstance(values, str):
            values = values.split(delimiter)

        for i, v in enumerate(values):
            v = str(v)
            values[i] = v.replace(",", "\\,")

        self.rows.append(values)
        self.num_rows += 1

    def error(self, message=''):
        """
        Exit the sampler with an error status and display the message parameter.
        **ERROR** *message*.  If `display_on_exit` is True then also
        output the contents of the sampler.  The script will
        exit with a status of 1

        :param message: The message to display after *ERROR*
        """
        self.add_headline("samplingStatus", "ERROR " + message)
        self._display()
        sys.exit(1)

    def warning(self, message=''):
        """
        Display the message parameter after a
        **WARN** status.  If `display_on_exit` is True then also
        output the contents of the sampler.

        :param message: The message to display after *WARN*
        """
        self.add_headline("samplingStatus", "WARN " + message)
        self._display()

    def ok(self, message=''):
        """
        Display the message parameter after an
        **OK** status.  If `display_on_exit` is True then also
        output the contents of the sampler.

        :param message: The message to display after *OK*
        """
        self.add_headline("samplingStatus", "OK " + message)
        self._display()

    def _display(self):
        if not self._display_on_exit:
            return

        if self._filename:
            fp = open(self._filename, "w")
        else:
            fp = sys.stdout

        print(",".join(self.headings), file=fp)
        for h in self.headlines:
            print("<!>{},{}".format(h, self.headlines[h]), file=fp)
        for r in self.rows:
            print(",".join(r), file=fp)




__author__ = "Andrew Lister"
__author_email__ = "andrew.lister@outlook.co.id"

import os
from pathlib import Path

# Set __doc__ to the first paragraph of README.md
_readme_path = Path(__file__).resolve().parent.parent / "README.md"
if _readme_path.exists():
    with _readme_path.open(encoding="utf-8") as f:
        lines = f.read().splitlines()

    first_paragraph = []
    for line in lines:
        if not line.strip():
            break
        first_paragraph.append(line)

    if first_paragraph:
        __doc__ = "\n".join(first_paragraph)


# Copyright © 2019 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# Provides utilities for querying the environment when an alert
# or user assignemnt event is generated. Each environment variable
# is placed in a member of the class allowing for consistent
# retrieval and parsing
#
# Dataview columns are also placed in a dict:
#    self.dataview_columns["cpu"]
# to allow alert enrichment

import os


class Environment:
    def __init__(self):
        """
        Populate the `Environment` object with commonly found
        environment variables from [ITRS geneos actions](https://docs.itrsgroup.com/docs/geneos/7.2.0/processing/monitoring-and-alerts/geneos_rulesactionsalerts_tr/index.html#script-actions).
        Useful when writing actions and triggers.  Please refer to the
        source code to see what is set.
        """

        self.location = os.getenv("location")
        """The value of the environment variable `location` - the `location` attribute"""
        self.application = os.getenv("application")
        """The value of the environment variable `application` - the `application` attribute"""
        self.environment = os.getenv("environment")
        """The value of the environment variable `environment` - the `environment` attribute"""
        self.variable = os.getenv("_VARIABLE")
        """The value of the environment variable `_VARIABLE` - Short name of the data-item
        if it is a managed variable, in the form <!>name for headlines or row.col for table
         ells. This value is provided for backwards compatibility."""
        self.action = os.getenv("_ACTION")
        """The value of the environment variable `_ACTION` - The name of the action being triggered"""
        self.value = os.getenv("_VALUE")
        """The value of the environment variable `_VALUE` -
        The value of the dataview cell the data-item belongs to (if any)."""
        self.managed_entity = os.getenv("_MANAGED_ENTITY")
        """The value of the environment variable `_MANAGED_ENTITY` -
        The name of the managed entity the data-item belongs to (if any)."""
        self.sampler = os.getenv("_SAMPLER")
        """The value of the environment variable `_SAMPLER` -
        The name of the sampler the data-item belongs to (if any)."""
        self.sampler_type = os.getenv("_SAMPLER_TYPE")
        """The value of the environment variable `_SAMPLER_TYPE` -
        The type of the sampler the data-item belongs to (if any)."""
        self.sampler_group = os.getenv("_SAMPLER_GROUP")
        """The value of the environment variable `_SAMPLER_GROUP` -
        The group of the sampler the data-item belongs to (if any)."""
        self.gateway = os.getenv("_GATEWAY")
        """The value of the environment variable `_GATEWAY` - 
        The name of the gateway firing the action."""
        self.rowname = os.getenv("_ROWNAME")
        """The value of the environment variable `_ROWNAME` - 
        The row name of the dataview cell the data-item belongs to (if any).  
        If it is a headline variable, then it is set to `_HEADLINE`"""
        if not self.rowname:
            self.rowname = os.getenv("_HEADLINE")
        self.column = os.getenv("_COLUMN")
        """The value of the environment variable `_COLUMN` -
        The column name of the dataview cell the data-item belongs to (if any)."""
        self.first_column = os.getenv("_FIRSTCOLUMN")
        """The value of the environment variable `_FIRSTCOLUMN` - 
        The name of the first column of the dataview the data-item belongs to (if any)."""
        self.dataview = os.getenv("_DATAVIEW")
        """The value of the environment variable `_DATAVIEW` - 
        The name of the dataview the data-item belongs to (if any)."""
        self.plugin_name = os.getenv("_PLUGINNAME")
        """The value of the environment variable `_PLUGINNAME` -
        The plugin name of the sampler the data-item belongs to (if any)."""
        self.rule = os.getenv("_RULE")
        """The value of the environment variable `_RULE` -
        The rule that triggered this action. This is the full path 
        to the rule including the rule groups i.e. group1 > group 2 > rulename"""
        self.host = os.getenv("_NETPROBE_HOST")
        """The value of the environment variable `_NETPROBE_HOST` -
        The hostname of the probe the data-item belongs to (if any). """
        self.probe = os.getenv("_PROBE")
        """The value of the environment variable `_PROBE` -
        The name of the probe the data-item belongs to (if any)."""
        self.kba_urls = os.getenv("_KBA_URLS")
        """The value of the environment variable `_KBA_URLS` -
        A list of application knowledge base article URLs, separated by newlines."""
        self.severity = os.getenv("_SEVERITY") or "critical"
        """The value of the environment variable `_SEVERITY` -
        The data-item severity. One of UNDEFINED, OK, WARNING, CRITICAL or USER.
        Converted to lower case"""
        self.severity = self.severity.lower()
        self.path = os.getenv("_VARIABLEPATH")
        """The value of the environment variable `_VARIABLEPATH` - The full gateway path 
        to the data-item."""
        self.assignee_email = os.getenv("_ASSIGNEE_EMAIL")
        """The value of the environment variable `_ASSIGNEE_EMAIL` -
        Name of the Geneos user assigned this item. The name is taken 
        from the user definition in the authentication section."""
        self.assignee_name = os.getenv("_ASSIGNEE_USERNAME")
        """The value of the environment variable `_ASSIGNEE_USERNAME` -
        Email address of the Geneos user assigned this item. The address is taken from
        the user definition in the authentication section."""
        self.assigner_name = os.getenv("_ASSIGNER_USERNAME")
        """The value of the environment variable `_ASSIGNER_USERNAME` -
        Name of the Geneos user assigning this item to another. The name is taken from
        the user definition in the authentication section."""
        self.comment = os.getenv("_COMMENT")
        """The value of the environment variable `_COMMENT` -
        Comment entered by the assigner or the user who unassigned the item. If no 
        comment is provided, this variable is set to a blank string ("")."""
        self.previous_comment = os.getenv("_PREVIOUS_COMMENT")
        """The value of the environment variable `_PREVIOUS_COMMENT` -
        Contents of the _COMMENT environment variable from the previous 
        assign/unassign event. """
        self.period_type = os.getenv("_PERIOD_TYPE")
        """The value of the environment variable `_PERIOD_TYPE` - 
        Period for which the item is assigned:
        * **Manual** — Assigned to the user until unassigned.
        * **UntilOk** — Assigned until severity is OK.
        * **Until severity changes to specified** — Assigned until severity changed to one specified by assigner.
        * **Until severity changes from specified** — Assigned until severity changed from specified by assigner.
        * **Until a specific date / time** — Assigned until a specific date / time.
        * **Until severity changes to specified or until a specific date / time** — Assigned until severity changed to one specified by assigner or a specific date / time.
"""

        self.dataview_columns = {}
        """A dictionary of all the columns found on the current dataview. These are
        environment variables starting with `_` and followed by a lowercase
        letter"""
        # Pull out all the columns from the dataview (only works if starts
        # with '_' and contains a lowercase letter
        for e in sorted(os.environ):
            if not e.startswith("_"):
                continue
            if e[1:] == self.first_column:
                continue
            if any(c for c in e if c.islower()):
                self.dataview_columns[e[1:]] = os.environ[e]



# Copyright © 2019-2025 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# Provides ALXdatabase class for executing SQL queries on MariaDB/MySQL,
# PostgreSQL, and SQLite backends using a consistent API.
#
from alx.app import ALXapp
import re
from typing import Any
from alx.strings import normalize
from sql_formatter.core import format_sql
# Only want to support modules that are installed.
try:
    import psycopg2
    has_postgres = True
except ImportError:
    psycopg2 = None
    has_postgres = False
try:
    import sqlite3
    has_sqlite = True
except ImportError:
    sqlite3 = None
    has_sqlite = False  # basically always True, part of stdlib
try:
    import mariadb as mysql
    has_mysql = True
except ImportError:
    try:
        import pymysql as mysql
        has_mysql = True
    except ImportError:
        has_mysql = False
        mysql = None


class ALXdatabase:
    def __init__(self, dbtype: str = 'mysql', user: str = None,
                 password: str = None, host: str = 'localhost', database: str = None,
                 port: int = 3306, autoconnect: bool = False,
                 autocommit: bool = False) -> None:
        """
        Simplifies and removes repetitive statements to connect to a database.

        :param dbtype: The database type.  Default is *mysql* which can also
         be used for *mariadb*. Supported options are
           * `mysql` -> `pip install pymysql`
           * `mariadb` -> `pip install mariadb`
           * `postgres` -> `pip install psycopg2`
           * `sqlite`
        :param user: The username to use
        :param password: The password to use
        :param host: The host to connect (default is `localhost`)
        :param database: The name of the database
        :param port: The port (default is mariadb, 3306)
        :param autoconnect: If True then connect to the database after
            initialisation. Default is False
        :param autocommit: If True then execute `commit` after each
            successful transaction

        You can also do this:
        ```
        with ALXdatabase(dbtype='sqlite', database=':memory:', autoconnect=True) as db:
            db.run("CREATE TABLE test (id INT)")
            db.run("INSERT INTO test (id) VALUES (1)")
        # <-- commits automatically, or rolls back if an error occurred
        ```
        """

        self.logger = ALXapp.logger
        """The default logger from the alx.app.ALXapp.logger. This class needs
        to be used so the logger is initialised"""
        self.cursor = None
        """The cursor assigned in `ALXdatabase.connect` after
        making the database connection"""
        self.connection = None
        """The connection assigned in `ALXdatabase.connect` after
        making the database connection"""
        self.autocommit = autocommit
        """Whether to automatically commit a successful transaction"""

        dbtype = normalize(dbtype)

        if dbtype == 'mariadb':
            self.dbtype = 'mysql'
        else:
            self.dbtype = dbtype

        self._params = {'user': user, 'password': password,
                        'host': host, 'database': database,
                        'port': port}
        if self.dbtype == 'sqlite':
            self._params['database'] = database or ':memory:'

        self.logger.info("Initialising %s database connection to %s on %s as %s",
                         dbtype, self._params['database'], self._params['host'],
                         self._params['user'])

        if autoconnect:
            self.connect()

    def connect(self) -> Any:
        """
        Initiates a connection to the database with parameters set
        in `ALXdatabase` constructor

        :return: The cursor from the connection made in `mariadb.connect`
        with the parameters set in `ALXdatabase`
        """
        try:
            if self.dbtype == 'mysql':
                if not has_mysql:
                    raise RuntimeError("MySQL/MariaDB support not available")
                self.connection = mysql.connect(**self._params)
            elif self.dbtype == 'postgres':
                if not has_postgres:
                    raise RuntimeError("PostgreSQL support not available")
                self.connection = psycopg2.connect(**self._params)
            elif self.dbtype == 'sqlite':
                if not has_sqlite:
                    raise RuntimeError("SQLite support not available")
                self.connection = sqlite3.connect(self._params['database'])
            else:
                raise ValueError(f"Unsupported database type: {self.dbtype}")
        except Exception as e:
            self.logger.error("Connection failed: %s", format(e))

        self.cursor = self.connection.cursor()
        self.logger.info("Connected successfully")
        return self.cursor

    def _convert_placeholders(self, sql: str) -> str:
        if self.dbtype == 'sqlite':
            return re.sub(r'%s', '?', sql)
        return sql

    def run(self, sql: str, name: str = None,
            params=None,
            multi: bool = False) -> list:
        """
        Tidies up the SQL string passed, logs the statement to
        `ALXapp.logger` and executes the statement on the
        `ALXdatabase` object.

        Supports parameterized queries using the DB-API parameter
        style (? or %s).

        If the statement is a *select*, then the result set is
        returned and *None* otherwise

        :param sql: The SQL statement to execute.
        :param name: Optionally name the query to identify it in the log
        :param params: A tuple or list of parameters to use with the SQL query.
        :param multi: If True, use executemany() for bulk inserts.

        :return: If a *select* statement then the result set
        from the call to execute on the `cursor` or
        an empty list if an `insert`, `update`, `upsert` or
        `replace` statement
        """
        sql = sql.strip()
        sql = self._convert_placeholders(sql)

        if name:
            log = "%s:\n%s" % (name, format_sql(sql))
        else:
            log = format_sql(sql)
        self.logger.debug(log)

        try:
            if multi and params:
                self.cursor.executemany(sql, params)
            elif params:
                self.cursor.execute(sql, params)
            else:
                self.cursor.execute(sql)
            if self.autocommit:
                try:
                    self.connection.commit()
                except Exception as e:
                    self.logger.warning("Autocommit failed: %s", e)
        except Exception as e:
            self.logger.error('SQL execution failed: %s', e)
            raise

        if sql.lower().startswith("select"):
            self.logger.info("%d rows returned", self.cursor.rowcount)
            return self.cursor.fetchall()

        self.logger.info("%d rows affected", self.cursor.rowcount)
        return []

    def commit(self):
        """
        Commit the current transaction.  This function should be called
        after modifying or inserting data. It is not done automatically
        to allow for exception handling

        :return: None
        """
        self.connection.commit()

    def rollback(self):
        """
        Rollback the current transaction. Do not commit the outstanding
        data. It is not done automatically to allow for exception handling

        :return: None
        """
        self.connection.rollback()

    def close(self) -> None:
        """
        Close the `ALXdatabase` connection and cursor and
        set them to None
        """
        try:
            if self.connection:
                self.connection.close()
            if self.cursor:
                self.cursor.close()
        except (sqlite3.Error, Exception):
            pass

        self.cursor = None
        self.connection = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if exc_type:
            self.rollback()
        else:
            self.commit()
        self.close()

    def __del__(self):
        try:
            self.close()
        except (sqlite3.Error, Exception):
            pass



# Copyright © 2019 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
#
# Parses and handles ITRS Geneos alerts, creating am html
# table suitable for sending via email or other alert mechanism
#

from alx.html import ALXhtml
from alx.strings import date_subst
from alx.itrs.environment import Environment
from alx.app import ALXapp
import json


class HtmlAlert:
    def __init__(self, environment: Environment):
        """
        Create an HTML table of information suitable for an alert.
        This alert can be sent using the ALXmail module.  Currently,
        it is used for email alerts and user assignment events

        :param environment: The environment created in alx.itrs.environment.Environment
        """
        self.environment = environment
        """Stores the environment passed"""
        self.config = ALXapp.read_lib_config(self)
        """The module configuration read from `alx.ini`"""
        self.colours = json.loads(self.config.get("alert", "colours"))
        """The alert colours read from `alx.ini`"""
        self.date_format = self.config.get("alert", "date_format")
        """The alert date format read from `alx.ini`"""
        self.style = "background-color: %s;" % self.colours[environment.severity]
        """The additional styles for the cells coloured with the severity from colours"""

    def create(self) -> str:
        """
        Populate the HTML table with the values in the environment created
        in a call to alx.itrs.environment.Environment
        :return: the formatted HTML table
        """
        e = self.environment
        html = ALXhtml()

        html.add_table()
        html.add_headings(["Variable", "Value"])
        html.start_row()
        html.add_cell("Severity")
        html.add_cell(e.severity.title(), style=self.style)
        html.end_row()
        html.add_row(["Gateway", e.gateway])
        html.add_row(["Application", e.application])
        html.add_row(["Location", e.location])
        html.add_row(["Host", e.host])
        html.add_row(["Date", date_subst(self.date_format)])
        html.add_row(["Managed Entity", e.managed_entity])
        html.add_row(["Sampler", e.sampler])
        html.add_headings(["Dataview", e.dataview])
        html.start_row()
        html.add_cell("Value")
        html.add_cell("%s %s is %s" % (e.rowname, e.column, e.value),
                      style=self.style)
        html.end_row()

        if len(e.dataview_columns) > 0:
            html.add_headings(["Dataview Columns", ""])

            for c in e.dataview_columns:
                if e.dataview_columns[c]:
                    html.add_row([c, e.dataview_columns[c]])

        html.end_table()

        return html.body + "</body>\n"

    


# Copyright © 2019-2025 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# Defines the ALXapp class and supporting infrastructure for application
# initialization, configuration, command-line parsing, and logging.

import configparser
import os
import sys
import argparse
from cryptography.fernet import Fernet
import json
import logging
import platform
from logging.handlers import TimedRotatingFileHandler
from collections import OrderedDict


class Paths:
    def __init__(self, appname: str, inifile: str = None) -> None:
        """
        A class to hold the default paths for the application

        If the `data` is needed, it must be created manually to avoid unnecessary
        data directories.

        :param appname: The application name
        :param inifile: If the inifile name is different from the default `app.ini` then it can be set here
        """

        basename = os.path.basename(os.path.dirname(sys.argv[0]))
        dirname = os.path.dirname(sys.argv[0])
        self.root = os.path.abspath(os.path.join(dirname, "..", ".."))
        """The root of the installation, perhaps `/opt/local/env` 
        (where *env* is *prod*, *test* or *dev*). Paths are calculated
         dynamically, based on file location"""
        self.bin = os.path.join(self.root, "bin")
        """The location of the start script: `root/bin`"""
        self.data = os.path.join(self.root, "data", basename)
        """The location of the application data: `root/data/app`"""
        self.etc = os.path.join(self.root, 'etc')
        """The location of the configuration files: `root/etc`"""
        self.log = os.path.join(self.root, "log", basename)
        """The location of the log files: `root/log/app`"""
        self.logfile = os.path.join(self.log, appname + ".log")
        """The log filename: `root/log/app/app.log`"""
        self.scripts = os.path.join(self.root, "scripts")
        """The location of the scripts: `root/scripts`"""
        self.top = os.path.join(self.root, "scripts", basename)
        """The location of the application scripts: `root/scripts/app`"""
        self.module_config_dir = self.get_module_config_dir()
        """The location of the module ini, key and env files"""
        self.keyfile = os.path.join(self.module_config_dir, "key")
        """The file from where to obtain the key: `~/.config/alx/key` or
        `%APPDATA%\\alx\\key` on Windows"""
        self.local_env = os.path.join(self.module_config_dir, "env")
        """The path to the venv in use: `~/.config/alx/env`"""
        self.local_config = os.path.join(self.module_config_dir, "alx.ini")
        """The path to the local configuration: `~/.config/alx/alx.ini`"""
        self.global_config = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "alx.ini")
        """The path to the module configuration: `<venv>/lib/.../site-packages/alx/alx.ini`"""

        self.config = os.path.join(self.etc, appname + ".ini")
        """The name of the config file determined from `self.etc`/app"""
        if inifile:
            self.config = os.path.join(self.etc, inifile)

    @staticmethod
    def get_module_config_dir():
        if platform.system() == "Windows":
            return os.path.join(os.environ['APPDATA'], "alx")
        else:
            return os.path.join(os.path.expanduser("~"), ".config", "alx")

    def __str__(self) -> str:
        return "\n".join(f"{k}: {v}" for k, v in vars(self).items() if isinstance(v, str))


class ALXapp:
    logger = logging.getLogger(os.path.splitext(os.path.basename(sys.argv[0]))[0])
    """The default logger used by all applications using `ALXapp`"""
    def __init__(self, description: str = "Unknown App",
                 args: list = None, appname: str = None,
                 inifile: str = None, epilog: str = None):
        """
        Initialise the `ALXapp` object which does a number of things:
        * Creates the application name from `sys.argv[0]` and stores it in
        `name` if not passed as a parameter
        * Adds a `--env` / `-e` argument and sets `self.environment`
        * Initialises the `Paths` class.
        * Reads the arguments provided in the `args` parameter.
        * Reads and parses the configuration file and stores the values in the ALXAppp object
        * Reads and parses the library configuration stored in `alx.ini`
        * Initialises and starts logging to `Paths.logfile`

        An example `alx.ini` or `$HOME/config/alx/alx.ini` or
        `%APPDATA%\alx` on Windows
        ```
        [DEFAULT]

        [logging]
        format:     %%(asctime)s: %%(levelname)s %%(module)s.%%(funcName)s:%%(lineno)d [%%(threadName)s] %%(message)s
        loglevel:   INFO
        days:       7
        maxsize:    10485760
        when:       midnight

        [mail]
        server:     mailhost
        from:       Super User <root@localhost>


        [html]
        css:
                p, h1, h2, h3, h4, h5, ul, ol, li {
                    font-family: Calibri, Arial, Helvetica
                }
                table {
                  font-family: "Trebuchet MS", Arial, Helvetica, sans-serif;
                  border-collapse: collapse;
                  width: 100%%;
                }
                td, th {
                  border: 1px solid #ddd;
                  padding: 5px;
                }
                tr:nth-child(even){background-color: #f2f2f2;}
                tr:hover {background-color: #ddd;}
                th {
                  padding-top: 8px;
                  padding-bottom: 8px;
                  text-align: left;
                  background-color: #0771af;
                  color: white;
                }
        ```
        :param description: A short description for the app - used with `--help`
        :param args: A list of arguments added with argparse.ArgumentParser.add_argument.  Example
        ```
        args = [
            ["-d", "--date", {"help": "store as '%%Y-%%m-%%d' date in database"}],
            ["-s", "--start", {"default": None, "type": str,
                               "help": "The first date from which to retrieve prices"}],
            ["-g", "--gui", {"action": "store_true", "default": False,
                             "help": "Display the browser."}]
        ]
        ```
        but you can just use this trimmed down version to store the
        arguments as strings
        ```
        args = [
            ["--date"],
            ["--start"],
         ]
        ```
        :param appname: The name of the application.  Default is the
         name of argv[0]
        :param inifile: The name of the configuration file.  Default
         is to create it from appname
        :param epilog: The text to use at the end of the help
         message when the app is called with `--help`
        """

        self.name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        """The name of the application from the parameter or calculated from `sys.argv`"""
        if appname:
            self.name = appname

        parser = argparse.ArgumentParser(description=description,
                                         epilog=epilog,
                                         formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        parser.add_argument("-e", "--env", default='dev',
                            help="Runtime Environment, dev, test or prod")

        if args:
            for a in args:
                if type(a[-1]) is dict:
                    arg = a[:-1]
                    kwargs = a[-1]
                    parser.add_argument(*arg, **kwargs)
                else:
                    parser.add_argument(*a)

        self.arguments = parser.parse_args()
        """A namespace of the arguments as parsed by 
        `argparse.Arguments.parse_args`"""
        self.environment = 'dev'
        """The execution environment - default is `dev` but any of 
        * test || uat || tst: **test**
        * prod || production || prd: **production**
        
        can be used to set *prod* or *test*
        """
        if self.arguments.env in ('test', 'uat', 'tst'):
            self.environment = 'test'
        elif self.arguments.env in ('prd', 'prod', 'production'):
            self.environment = 'prod'

        self.paths = Paths(self.name, inifile)
        """The `Paths` namespace that holds path information"""

        self.config = self.read_config(self.paths.config)
        """The application configuration read from `Paths.config` from a 
        call to `ALXApp.read_config`. The configuration values are assigned
        to the `ALXApp` class with a call to `ALXApp.parse_config`"""
        if self.config and self.environment in self.config:
            self.parse_config(self, self.config[self.environment])

        self.key = None
        """The key to encrypt and decrypt encoded strings"""

        self.libconfig = self.read_lib_config()
        """The global library configuration from `alx.ini`"""

        # Start logging with all configured parameters
        self.start_logging()

    @staticmethod
    def parse_config(obj: object, config: configparser.SectionProxy):
        """
        Parses the `config` and stores in `obj` as the appropriate type.
        It works out if it is a boolean, float, integer or string. If
        the configuration value starts with a `[` or `{` then it is
        loaded using `json.loads` and stored in an `OrderedDict`

        The config file should be structured so there are sections named
        * `[DEFAULT]`
        * `[dev]`
        * `[test]`
        * `[prod]`
        This allows settings to be overridden simply by specifying
        `-e env`. If a setting is not specified in an environment
        section, then the value from `[DEFAULT]` is used. For example
        ```
        [DEFAULT]
        loglevel:   INFO

        [dev]
        loglevel:   DEBUG

        [test]
        loglevel:   INFO

        [prod]
        loglevel:   WARN
        ```
        All the config file elements are stored in the `ALXapp` object.
        If you define:
        ```
        foo: bar
        bar: 1
        baz: [1, 2, 3, 4]
        ```
        in the ini file, then your `ALXapp` object, say `app`, will have
        these elements defined (either from the `[DEFAULT]` section or the
        appropriate environment:
        ```
        app.foo = "bar"
        app.bar = 1
        app.baz = [1, 2, 3, 4]
        ```
        Additionally, if you specify `$data` in the value, it will be
        expanded to the data directory of the application.  For example
        ```
        output_file = $data/output.txt
        ```
        will result in:
        ```
        app.output_file = /opt/local/prod/data/app/output.txt
        ```

        :param obj: The object in which to store the configuration
         values (usually `self`)
        :param config: The `configparser` object.  It can be a section
        like `config[app.environment]`
        """

        try:
            for item in config:
                # Add all the config values as string values in the app
                value = config.get(item)
                if '$data' in value:
                    if hasattr(obj, 'paths'):
                        value = value.replace('$data', obj.paths.data)
                        setattr(obj, item, value)
                    else:
                        continue
                elif value in ('True', 'False', 'true', 'false'):
                    # Convert to boolean
                    setattr(obj, item, config.getboolean(item))
                else:
                    try:
                        # Is it an integer?
                        i = int(value)
                        setattr(obj, item, i)
                    except (ValueError, TypeError):
                        try:
                            # or a float?
                            f = float(value)
                            setattr(obj, item, f)
                        except (ValueError, TypeError):
                            # Only a string left....
                            if value.startswith('[') or value.startswith('{'):
                                try:
                                    setattr(obj, item, json.loads(
                                        value, object_pairs_hook=OrderedDict))  # type: ignore[arg-type]
                                except json.JSONDecodeError:
                                    setattr(obj, item, value)
                            else:
                                setattr(obj, item, value)
        except Exception:
            raise

    @staticmethod
    def read_config(filename: str, filename2: str = None):
        """
        Reads the configuration file in `filename` using the parser
        passed in `parser` which is typically the default

        :param filename: Name of the configuration file to read
        :param filename2: Name of a second config file to override first
        :return: The `configparser.ConfigParser` configuration
        """
        if not os.path.isfile(filename):
            return None

        config = configparser.ConfigParser()
        if filename2:
            config.read([filename, filename2])
        else:
            config.read(filename)

        return config

    def _create_configuration_files(self) -> None:
        """
        Create some initial configuration files

        :return: None
        """
        os.makedirs(self.paths.module_config_dir, exist_ok=True)

        if not os.path.isfile(self.paths.local_config):
            print("*** NOTE: Creating user config file '{}'".format(self.paths.local_config))
            gc = ALXapp.read_config(self.paths.global_config)
            with open(self.paths.local_config, 'w') as f:
                for s in gc.sections():
                    f.write("[%s]\n\n" % s)

        if not os.path.isfile(self.paths.keyfile):
            print("*** NOTE: Creating key file '{}'".format(self.paths.keyfile))
            # Create an initial fernet key.  User to adjust as required
            with open(self.paths.keyfile, 'w') as f:
                f.write("%s\n" % Fernet.generate_key().decode())
            # Ensure strict permissions
            os.chmod(self.paths.keyfile, 0o600)

        if not os.path.isfile(self.paths.local_env):
            print("*** NOTE: Creating local environment file: '{}'".format(self.paths.local_env))
            # Assume that the current python invocation is the path to the virtual env
            with open(self.paths.local_env, 'w') as f:
                venv = os.path.dirname(os.path.dirname(sys.executable))
                f.write("venv=%s\n" % venv)

    def read_lib_config(self) -> configparser.ConfigParser:
        """
        Reads and parses the module configuration file and the local
        configuration file (`~/.config/alx/alx.ini` or
        `%APPDATA%/alx/alx.ini`). The local configuration overrides
        any values in the module configuration file.

        On first invocation, the user files are checked and created if
        they don't exist.

        :return: The `configparser.ConfigParser` configuration
        """
        if not hasattr(self, 'paths'):
            self.paths = Paths("alx-common")

        # If the user configuration doesn't exist, create them
        if not all([
            os.path.isdir(self.paths.module_config_dir),
            os.path.isfile(self.paths.local_config),
            os.path.isfile(self.paths.keyfile),
            os.path.isfile(self.paths.local_env),]):
            self._create_configuration_files()

        merged_config = ALXapp.read_config(self.paths.global_config,
                                           self.paths.local_config)

        return merged_config

    def start_logging(self):
        """
        Sets up a standard logger using the configuration in `alx.ini`
        By default, log files are rolled at midnight and kept for the
        number of days configured in `alx.ini`.  The loglevel and
        format are also set in the config file.

        If the application is in development mode, a console logger
        is also added for convenience
        """
        days = self.libconfig.getint('logging', 'days')
        log_format = self.libconfig.get('logging', 'format')
        when = self.libconfig.get('logging', 'when')

        if hasattr(self, 'loglevel'):
            loglevel = self.loglevel
        else:
            loglevel = self.libconfig.get('logging', 'loglevel')

        self.logger.setLevel(loglevel)

        if not os.path.isdir(self.paths.log):
            os.makedirs(self.paths.log)

        fh = TimedRotatingFileHandler(self.paths.logfile, when=when,
                                      backupCount=days)
        formatter = logging.Formatter(log_format)
        fh.setLevel(loglevel)
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        if self.is_dev():
            # Also log to console...
            ch = logging.StreamHandler()
            ch.setLevel(loglevel)
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)

        self.logger.debug("Starting application '%s', logging at level '%s'",
                          format(self.name), format(loglevel))

    def set_log_level(self, level: str) -> None:
        """
        Change the logging level at runtime

        :param level: Log level string ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        """
        self.logger.setLevel(level)
        # Update all handlers to the new level
        for handler in self.logger.handlers:
            handler.setLevel(level)

        self.logger.info(f"Log level changed to {level}")

    def _read_key(self) -> Fernet:
        """Reads the key from `~/.config/alx/key` and uses it to encrypt and
         decrypt strings using the `cryptography` python module"""

        if not os.path.exists(self.paths.keyfile):
            self.logger.error("Could not open %s", self.keyfile)
            sys.exit(1)
        st = os.stat(self.paths.keyfile)
        if platform.system() != 'Windows' and int(oct(st.st_mode)[3:]) > 600:
            # Apologies to Windows users but your security is too messy
            self.logger.error("Check permissions on %s.  Too open", self.keyfile)
            sys.exit(1)

        with open(self.paths.keyfile) as k:
            key = k.read().strip()

        fernet = Fernet(key)

        return fernet

    def encrypt(self, string: str) -> str:
        """
        Encrypts the password using the key read from `~/.config/alx/key`. This
        function provides an easy way to avoid storing plain text passwords

        The keyfile needs to have permission `0600` and look something like
        ```
        P9H7kabcdefghijwlSrfVnKKqV5VCnW5RE8OT21eH5k=
        ```
        Please refer to `decrypt` for how to create a key.

        Note that due to the inconsistent implementation of permissions
        on Windows, this check is ignored

        :param string: The string to encrypt.
        :return: The encrypted string
        """
        if not self.key:
            self.key = self._read_key()
        encoded = self.key.encrypt(string.encode())
        return encoded.decode()

    def decrypt(self, string: str) -> str:
        """
        Decrypts the password using the key read from `~/.config/alx/key`. This
        function provides an easy way to avoid storing plain text passwords

        Your code will then look like:
        ```
        self.passwd = app.decrypt(app.config.get('mysql', 'password'))
        self.user = app.decrypt(app.config.get('mysql', 'user'))
        ```
        To generate a key, use this one-liner and place the ouput in `~/.config/alx/key`:
        ```
        python -c "from cryptography.fernet import Fernet; print(Fernet.generate_key().decode())"`
        ```

        If there are different users of an application using `ALXapp` then
        the key must be consistent or the string will fail to decrypt.

        :param string: The string to decrypt.
        :return: The decrypted string
        """
        if not self.key:
            self.key = self._read_key()
        decoded = self.key.decrypt(string.encode()).decode()
        return decoded

    def is_dev(self) -> bool:
        """
        :return: `True` if running in dev mode and `False` otherwise
        """
        return self.environment == 'dev'

    def is_test(self) -> bool:
        """
        :return: `True` if running in test mode and `False` otherwise
        """
        return self.environment == 'test'

    def is_prod(self) -> bool:
        """
        :return: `True` if running in prod mode and `False` otherwise
        """
        return self.environment == 'prod'



# Copyright © 2019-2025 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# String utility functions including normalization, trimming, and
# conversion between naming styles (e.g., space to underscore).

import re
from datetime import datetime, timezone
from unidecode import unidecode


def date_subst(fmt: str, when: datetime = None,
               tz: timezone = timezone.utc) -> str:
    """
    Return a formatted string of the date passed date according to
    `strftime`.

    :param fmt: date format as for strftime()
    :param when: the date to convert - default is now.
    :param tz: the timezone of the date.  default is UTC
    :return: the date as a formatted string
    """
    if not when:
        when = datetime.now(tz).astimezone()

    return when.strftime(fmt)


def normalize(s: str) -> str:
    """
    Normalize a string: trim, lowercase, collapse internal whitespace.

    :param s: The input string
    :return: The normalized string
    """
    return " ".join(s.strip().lower().split())


def replace_accents(s: str) -> str:
    """
    Convert accented characters to ASCII equivalents.

    E.g., é → e, ñ → n, ü → u

    :param s: Input string
    :return: ASCII-only version of string
    """
    return unidecode(s)


def replace_spaces(s: str, c: str = ".") -> str:
    """
    Replace spaces in a string with another character. Default is '.'

    :param s: The input string
    :param c: The character to use in the substitution., Default is '.'

    :return: The new string
    """
    return s.replace(" ", c)


def sanitize_filename(s: str, c: str = ".") -> str:
    """
    Remove all characters that are not suitable for Unix filenames.
    Essentially, all characters not in the set  [^a-zA-Z0-9._-] but
    also:
    ```
    * accents are removed: é becomes e
    * "word - word" -> "word-word"
    * `c ` -> `c`
    * & -> and
    ```

    :param s: The input string
    :param c: The character to use for the replacement. Default is '.'
    :return: The new string
    """
    s = s.strip()
    s = replace_accents(s)                                          # ✅ Normalize accents
    s = re.sub(r'\s*-\s*', '-', s)                 # "word - word" -> "word-word"
    s = re.sub(r'[()[\]\'{}]', '', s)               # Remove brackets, parens, quotes, braces
    s = re.sub(r'&', 'and', s)                      # & -> and
    s = replace_spaces(s, c)                                    # Replace spaces with `c`
    s = re.sub(rf'[^{re.escape(c)}a-zA-Z0-9._-]', c, s)   # Replace everything else

    return s



# Copyright © 2019-2025 Andrew Lister
# License: GNU General Public License v3.0 (see LICENSE file)
#
# Description:
# Utility functions for rendering and escaping HTML content. Also includes
# the ALXhtml class for composing text and HTML-based email messages.

from alx.app import ALXapp
from typing import Any


class ALXhtml:
    def __init__(self, title: str = None) -> None:
        """
        An HTML module to simplify the creation of HTML through the use of
        methods.  This makes the code calling the library much more readable

        :param title: The title to use in the head section of the HTML
        """
        self.config = ALXapp.read_lib_config(self)
        """Adds the configuration read from `alx.ini`"""
        self.logger = ALXapp.logger
        """The default logger from the alx.app.ALXapp.logger"""
        self.css = "\n" + self.config.get('html', 'css') + "\n"
        """Stores the css read from the ini file.  More styles can be added with `set_css`"""
        self.head = "<head>\n"
        """Initialise the `head` with `title` if specified"""
        if title:
            self.head += "<title>" + title + "</title>\n"
        self.doctype = "<!DOCTYPE html>\n"
        """Initialise the DOCTYPE"""
        self.html = "<html>\n"
        """Initialise the `html` text"""
        self.body = "<body>\n"
        """Initialise the `body` text"""
        self._current_row = 0
        """Internal variable to hold current row number"""
        self._current_column = 0
        """Internal variable to hold current column number"""
        self.column_headings = []
        """Holds column indexes that should be treated as headings.
        For example, to mark the first column as a heading, set to 
        `[True]` through `set_column_headings`"""
        self.column_alignments = []
        """Holds column index alignments to treat alignments
        differently to the `css`. For example, to set the
         alignemnt of the first column to left, call 
         `set_column_alignments` with a parameter of `["left"]`"""

    def add_css(self, css: str) -> None:
        """
        Allow the default css to be overridden if necessary

        :param css: A block of css. The `style` tags should not be included
        """

        self.css = "\n" + css + "\n"

    def add_heading(self, number: int, heading: str) -> None:
        """
        adds a heading of level 'number'.  This method should likely
        be accessed by the helper functions - `add_h1`, `add_h2`, etc.

        :param number: The heading level
        :param heading: The text for the heading
        """
        self.body += "<h%d>" % number + heading + "</h%d>\n" % number

    def add_h1(self, heading: str) -> None:
        """
        Adds level 1 heading

        :param heading: The text for the header
        """
        self.add_heading(1, heading)

    def add_h2(self, heading: str) -> None:
        """
        Adds level 2 heading

        :param heading: The text for the header
        """
        self.add_heading(2, heading)

    def add_h3(self, heading: str) -> None:
        """
        Adds level 3 heading

        :param heading: The text for the header
        """
        self.add_heading(3, heading)

    def add_h4(self, heading: str) -> None:
        """
        Adds level 4 heading

        :param heading: The text for the header
        """
        self.add_heading(4, heading)

    def add_h5(self, heading: str) -> None:
        """
        Adds level 5 heading

        :param heading: The text for the header
        """
        self.add_heading(5, heading)

    def add_horizontal_line(self) -> None:
        """
        Adds a horizontal line
        """
        self.body += "<hr>\n"

    def add_paragraph(self, paragraph: str) -> None:
        """
        Adds a paragraph of text

        :param paragraph: The text for the paragraph
        """
        self.body += "<p>\n" + paragraph

    def add_bold_text(self, text: str) -> None:
        """
        Add some text in bold

        :param text: The text to add
        """
        self.body += "<b>" + text + "</b>"

    def add_italic_text(self, text: str) -> None:
        """
        Add some text in italics

        :param text: The text to add
        """
        self.body += "<i>" + text + "</i>"

    def add_bold_italic_text(self, text: str) -> None:
        """
        Add some text in bold italics

        :param text: The text to add
        """
        self.body += "<b><i>" + text + "</i></b>"

    def add_ul(self) -> None:
        """
        Adds an unordered list.  Finish with `end_ul`
        """
        self.body += "<ul>\n"

    def add_ol(self, tag: str = '1') -> None:
        """
        Adds an ordered list.  Finish with `end_ol`
        :param tag: The tag type:
        * "1": The list items will be numbered with numbers (default)
        * "A": The list items will be numbered with uppercase letters
        * "a": The list items will be numbered with lowercase letters
        * "I": The list items will be numbered with uppercase Roman numerals
        * "i": The list items will be numbered with lowercase Roman numerals
        """
        self.body += "<ol type='%s'>\n" % tag

    def add_item(self, item: str) -> None:
        """
        Adds an item to the ordered or unordered list

        :param item: The line item
        """
        self.body += "  <li>" + item + "</li>\n"

    def end_ol(self) -> None:
        """
        Ends the ordered list
        """
        self.body += "</ol>\n"

    def end_ul(self) -> None:
        """
        Ends the unordered list
        """
        self.body += "</ul>\n"

    def add_table(self, style: str = "") -> None:
        """
        Adds a table to the HTML

        :param style: Adds an optional style to the table.  Do not include the `style` tags
        """
        self.body += "<table"
        if style:
            self.body += " style='%s'" % style
        self.body += ">\n"
        self._current_row = self._current_column = 0

    def set_column_headings(self, headings: list = None) -> None:
        """
        Set columns to be headings (bold / different background) for each
        column where the list element is True

        :param headings: A list of boolean values marking a column as a
            heading.  By default, the first column is set
        :return: None
        """
        if not headings:
            headings = [True]

        self.column_headings = headings

    def set_column_alignments(self, alignments: list = None) -> None:
        """
        Set column alignments to the values in `alignments`

        :param alignments: A list of str values marking the justification of a
        column: "left", "center" or "right".  If a list element is not set
        then the `css` rules are followed. By default, the first column is
        set to left justified
        :return: None
        """
        if not alignments:
            alignments = ["left"]

        self.column_alignments = alignments

    def start_row(self, style: str = "") -> None:
        """
        Start a new row in the current table

        :param style: The css style to use.
        :return: None
        """
        if not style:
            self.body += "  <tr>\n"
        else:
            self.body += "  <tr style='%s'>\n" % style

    def add_cell(self, value: Any, tag: str = 'd', style: str = "") -> None:
        """
        Add a single cell to the current table

        :param value: This can be of any type, it will be converted to a string
        :param tag: Can be set to 'h' to create a heading
        :param style: A css style for the cell.  Do not include the `style` tags
        """
        if (self._current_column < len(self.column_headings) and
                self.column_headings[self._current_column]):
            # This column defined as a heading so mark as `th`
            tag = 'h'
        if (self._current_column < len(self.column_alignments) and
                self.column_alignments[self._current_column]):
            # This column has an alignment set, mark as so...
            style += ("text-align: %s; " %
                      self.column_alignments[self._current_column].lower())

        if not style:
            self.body += "    <t%s>%s</t%s>\n" % (tag, str(value), tag)
        else:
            self.body += "    <t%s style='%s'>%s</t%s>\n" % (tag, style, str(value), tag)
        self.logger.debug("Row: %d, Column: %d (%s)", self._current_row,
                          self._current_column, str(value))
        self._current_column += 1

    def add_cells(self, values: list, tag: str = 'd', style: str = "") -> None:
        """
        Add a list of cells, possibly to a partial row
        :param values: A list of values to add.  Elements can be of any type
        :param tag: The tag: either 'd' or 'h'. Default is 'd'
        :param style: A css style for the cells
        :return: None
        """
        for v in values:
            self.add_cell(v, tag, style)

    def end_row(self) -> None:
        """
        End the current row in the current table
        """
        self.body += "  </tr>\n"
        self._current_row += 1
        self._current_column = 0

    def add_row(self, values: list, tag: str = "d", style: str = ""):
        """
        Add a whole row to a table.  The values passed in should be a
        list of values that make up the complete row

        :param values: A list of values
        :param tag: The tag for the row - 'h' or 'd'
        :param style: a css style for the row (do not include `style` tags)
        """

        self.start_row(style)
        for value in values:
            self.add_cell(value, tag, style)
            # self.body += "    <t%s>" % tag + str(td) + "</t%s>\n" % tag
        self.end_row()

    def add_headings(self, values: list) -> None:
        """
        Calls `add_row` with a tag of `h` and passes `values` to
        :param values: a list of values to be used as headings
        """
        self.add_row(values, "h")

    def end_table(self) -> None:
        """
        End the current table with a `/table` tag.  Not ending the table can
        lead to unexpected results!
        """
        self.body += "</table>\n"

    def add_html(self, value: str) -> None:
        """
        Add raw HTML to the class

        :param value: The raw HTML
        """
        self.body += value + "\n"

    def add_url(self, target: str, text: str = None) -> None:
        """
        Add a url to the object that will be displayed as a clickable link
        :param target: The target address or URL
        :param text: The text to display for the link. If ommitted, the target
            name is used
        """
        if not text:
            text = target
        self.body += "<a href='%s'>%s</a>\n" % (target, text)

    def get_html(self) -> str:
        """
        Put all the elements together and return a formatted HTML document

        :return: The complete HTML document
        """
        value = self.doctype
        value += self.html
        value += self.head
        value += "<style>" + self.css + "\n</style>\n"
        value += "</head>\n"
        value += self.body
        value += "</body>\n"
        value += "</html>\n"

        return value
